/*  equation2.h
 *
 *  Copyright (C) 2010-2012 Andreas von Manteuffel
 *  Copyright (C) 2010-2012 Cedric Studerus
 *
 *  This file is part of the package Reduze 2.
 *  It is distributed under the GNU General Public License version 3
 *  (see the file GPL-3.0.txt or http://www.gnu.org/licenses/gpl-3.0.txt).
 */

/** Commutative Operators OP, OPPROD, OPSUM and the symbolic
 * integrals and equations INT2, Equation2 EquationList2.
 *
 *   OP(i)  INT2(a_1,...,a_i,...a_n)  =  INT2(a_1,...,a_i + 1,...a_n)
 *   OP(-i) INT2(a_1,...,a_i,...a_n)  =  INT2(a_1,...,a_i - 1,...a_n)
 *
 * with a_i from Z   and   INT2: Z^n --> R   and   OP(0) = identity
 */

#ifndef EQUATION2_H_
#define EQUATION2_H_

#include <ginac/ginac.h>
#include "equation.h"

class Propagator;

namespace Reduze {

class IntegralFamily;
class IdentityList;
class Equation2;
class EquationList2;

/// single shift operator OP(i)
class OP {
private:
	/// increases/decreases position i of the INT2(a1,....,an)
	// n = +1 increases a1, n = -1 decreases a1, one OP shifts only by one
	int n_;

public:
	/// standard constructor, represents the unity  OP(0)
	OP() :
		n_(0) {
	}
	/// constructor
	explicit OP(int n) :
		n_(n) {
	}

	int n() const {
		return n_;
	}

	bool operator==(const OP & other) const {
		return (n_ == other.n_);
	}
	bool operator!=(const OP & other) const {
		return (n_ != other.n_);
	}
	bool operator<(const OP & other) const {
		return (n_ < other.n_);
	}

	friend std::ostream & operator<<(std::ostream & strm, const OP & in);
	friend class OPPROD;
};

/// product of operators OP(i_1), ... , OP(i_k)
class OPPROD {
private:
	GiNaC::ex fac_;
	std::list<OP> prod_;

public:
	/// standard constructor gives the unity
	OPPROD() :
		fac_(1), prod_(1, OP()) {
	}
	explicit OPPROD(const OP & op) :
		fac_(1), prod_(1, op) {
	}
	OPPROD(const GiNaC::ex & e, const OP & op);

	size_t size() const {
		return prod_.size();
	}

	const GiNaC::ex& fac() const {
		return fac_;
	}

	/// sorts the products in ascending order and removes/replaces terms op*op^1 by the unity
	void unique();

	/// compares the size only
	bool operator<(const OPPROD & other) const;
	/// returns true if the product are equal up to the factor 'fac'
	bool operator==(const OPPROD & other) const;
	bool operator!=(const OPPROD & other) const;
	OPPROD & operator*=(const OPPROD & other);

	friend const OPPROD operator*(const OPPROD & a, const OPPROD & b);
	friend std::ostream & operator<<(std::ostream & strm, const OPPROD & in);
	friend class OPSUM;
	friend class INT2;
};

/// sum of instances of OPPROD
class OPSUM {
private:
	std::list<OPPROD> sum;
public:
	/// standard constructor gives the unity with prefactor 0
	OPSUM() {
	}
	explicit OPSUM(const OP & op) :
		sum(1, OPPROD(op)) {
		unique();
	}
	explicit OPSUM(const OPPROD & prod) :
		sum(1, prod) {
		unique();
	}

	size_t size() const {
		return sum.size();
	}
	void clear() {
		sum.clear();
	}

	/// sorts and uniques the individual OPPRODs
	void unique();
	/// collects together equal OPPRODs
	void collect();

	OPSUM & operator+=(const OPSUM & other);
	OPSUM & operator-=(const OPSUM & other);
	OPSUM & operator*=(const OPSUM & other);

	friend const OPSUM operator+(const OPSUM & a, const OPSUM & b);
	friend const OPSUM operator-(const OPSUM & a, const OPSUM & b);
	friend const OPSUM operator*(const OPSUM & a, const OPSUM & b);
	friend std::ostream & operator<<(std::ostream & strm, const OPSUM & in);
	friend class INT2;
};

class Equation2;

/// integral with general exponents
/** note: somewhat confusing: INT2 includes prefactor, INT does not **/
class INT2 {
private:
	/// prefactor
	GiNaC::ex fac_;
	/// the integer entries of the vector v are the exponents of the propagators
	GiNaC::exvector ep_;
	const IntegralFamily* integralfamily_;
public:
	/// standard constructor generates an integral with values a1, ..., an from PropagatorExponents
	INT2(const IntegralFamily* m);
	/// copy constructor
	INT2(const INT2 & in);
	INT2(const IntegralFamily* m, const GiNaC::exvector& v,
			const GiNaC::ex& fac);

	const IntegralFamily* integralfamily() const {
		return integralfamily_;
	}
	const GiNaC::exvector& v() const {
		return ep_;
	}
	const GiNaC::ex& fac() const {
		return fac_;
	}

	INT2 & operate(const OP & op);
	INT2 & operate(const OPPROD & prod);
	void operate(const OPSUM & sum, Equation2 & eq) const;

	friend std::ostream & operator<<(std::ostream & strm, const INT2 & in);

};

/// linear combination of integrals with generic exponents
/** \todo: Equation2 is actually only used as a generic linear combination
 *       => rename it accordingly (relevant for diff. eqns.) */
class Equation2 {
public:
	void append(const INT2 & in) {
		l.push_back(in);
	}
	void operate(const OPSUM & sum);
	LinearCombination apply(const INT&) const;

private:
	friend std::ostream & operator<<(std::ostream & strm, const Equation2 & in);
	std::list<INT2> l;
	friend class IdentityGenerator;

};

/// list of instances of Equation2
class EquationList2 {
public:
	size_t size() const {
		return l.size();
	}
	void append(const Equation2 & in) {
		l.push_back(in);
	}
	void save(const std::string & filename); // \todo not used

private:
	friend std::ostream & operator<<(std::ostream & strm,
			const EquationList2 & in);
	friend class IdentityGenerator;

	std::list<Equation2> l;
};

// functions

/// transforms combinations of Propagators to OPSUMs
OPSUM ex_to_OPSUM(const GiNaC::ex & e, const std::vector<Propagator>& props);

} // namespace Reduze


#endif /* EQUATION2_H_ */
