/*  identitygenerator.cpp
 *
 *  Copyright (C) 2010-2012 Andreas von Manteuffel
 *  Copyright (C) 2010-2012 Cedric Studerus
 *
 *  This file is part of the package Reduze 2.
 *  It is distributed under the GNU General Public License version 3
 *  (see the file GPL-3.0.txt or http://www.gnu.org/licenses/gpl-3.0.txt).
 */

#include "identitygenerator.h"
#include "equation2.h"
#include "kinematics.h"
#include "integralfamily.h"
#include "sectormappings.h"
#include "functions.h"
#include "files.h"
#include "equationlist.h"
#include "ginacutils.h"
#include "yamlutils.h"

using namespace std;
using namespace GiNaC;

namespace Reduze {

// register type
namespace {
YAMLProxy<IdentityGeneratorOptions> dummy;
}

// IdentityGeneratorOptions

IdentityGeneratorOptions::~IdentityGeneratorOptions() {
}

void IdentityGeneratorOptions::set_map_on_coeff(const GiNaC::exmap& m) {
	map_on_coeff_ = m;
}
void IdentityGeneratorOptions::set_transform_integrals_to_equivalent(bool b) {
	transform_integrals_to_equivalent_ = b;
	if (transform_integrals_to_equivalent_)
		map_on_integral_ = &SectorMappings::get_equivalent;
	else
		map_on_integral_ = 0; // has been set to non zero value in init()
}
void IdentityGeneratorOptions::set_solve_equations(bool b) {
	solve_equations_ = b;
}
void IdentityGeneratorOptions::set_subsectors_to_zero(bool b) {
	set_subsectors_to_zero_ = b;
	if (set_subsectors_to_zero_)
		eliminate_subsectors_by_shift_ = false;

}
void IdentityGeneratorOptions::set_eliminate_subsectors_by_shift(bool b) {
	eliminate_subsectors_by_shift_ = b;
	if (set_subsectors_to_zero_)
		eliminate_subsectors_by_shift_ = false;
}
void IdentityGeneratorOptions::init() {
	if (transform_integrals_to_equivalent_)
		map_on_integral_ = &SectorMappings::get_equivalent;
}

void IdentityGeneratorOptions::print(YAML::Emitter& os) const {
	using namespace YAML;
	os << BeginMap;
	os << Key << "transform_integrals_to_equivalent" << Value
			<< transform_integrals_to_equivalent_;
	//	os << Key << "solve_equations" << Value << solve_equations_;
	os << Key << "set_subsectors_to_zero" << Value << set_subsectors_to_zero_;
	os << Key << "eliminate_subsectors_by_shift" << Value
			<< eliminate_subsectors_by_shift_;
	os << Key << "map_on_coeff" << Value << Flow << map_on_coeff_;
	os << EndMap;
}

void IdentityGeneratorOptions::read(const YAML::Node& node) {
	verify_yaml_spec(node);
	const YAML::Node* n;
	if ((n = node.FindValue("transform_integrals_to_equivalent")))
		*n >> transform_integrals_to_equivalent_;
	//	if ((n = node.FindValue("solve_equations")))
	//		*n >> solve_equations_;
	if ((n = node.FindValue("set_subsectors_to_zero")))
		*n >> set_subsectors_to_zero_;
	if ((n = node.FindValue("eliminate_subsectors_by_shift")))
		*n >> eliminate_subsectors_by_shift_;
	if ((n = node.FindValue("map_on_coeff"))) {
		lst l = Files::instance()->kinematics()->kinematic_invariants();
		Reduze::read(*n, map_on_coeff_, l);
	}
	init();
}

// IdentityGenerator

IdentityGenerator::IdentityGenerator() {
}
IdentityGenerator::IdentityGenerator(const IdentityGeneratorOptions& options) :
	options_(options) {
}
IdentityGenerator::~IdentityGenerator() {
	while (!template_by_ic_.empty()) {
		delete template_by_ic_.begin()->second;
		template_by_ic_.erase(template_by_ic_.begin());
	}
}

void IdentityGenerator::generate(const INT& in, IdentityList& out,
		const string& type) {
	IdentityList identities;
	if (type == "sector_relations")
		genid_shift(in, identities);
	else if (type == "sector_symmetries")
		genid_symmetry(in, identities);
	else
		genid_template(in, identities, type);

	if (options_.eliminate_subsectors_by_shift_) {
		set<INT> ints;
		identities.find_subsector_INT(static_cast<int> (in.t()), ints);
		IdentityList replace;
		genid_shift(ints, replace);
		identities.replace(replace);
	}
	out.splice(identities);
}

void IdentityGenerator::generate(const set<INT>& s, IdentityList& out,
		const string& type) {
	set<INT>::const_iterator it;
	for (it = s.begin(); it != s.end(); ++it)
		generate(*it, out, type);
}

/// \todo: this implementation could use Equation2's conversion method
///        (which might be slightly slower than doing all in one go)
void IdentityGenerator::genid_template(const INT& in, IdentityList& result,
		const string& type) {
	const EquationList2& eqs = get_template_equation(in.integralfamily(), type);
	IdentityList identities;
	int8 in_t = in.t();
	unsigned num_props = in.integralfamily()->num_propagators();
	const exvector& prop_exps = in.integralfamily()->propagator_exponents();
	exmap m; // map to replace symbols by numbers;
	for (unsigned i = 0; i < num_props; i++)
		m[prop_exps[i]] = static_cast<int> (in.v_i(i));
	list<Equation2>::const_iterator it;
	for (it = eqs.l.begin(); it != eqs.l.end(); ++it) {
		Identity ident;
		list<INT2>::const_iterator it2;
		for (it2 = it->l.begin(); it2 != it->l.end(); ++it2) {
			ASSERT(it2->integralfamily() == in.integralfamily());
			INT newin(*it2, m);
			// sub-topologies just have a smaller t
			if (!options_.set_subsectors_to_zero_ || !(newin.t() < in_t)) {
				ex coeff = (it2->fac()).subs(m, subs_options::no_pattern);
				ident.insert(newin, coeff, false);
			}
		}
		identities.push_back(ident);
	}
	apply_remaining_options(identities);
	result.splice(identities);
}

Identity IdentityGenerator::get_shift_identity(const INT& i,
		bool set_subsecs_to_zero) {

	// map: crossing to minimal equivalent crossing
	const map<Crossing, Crossing>& crossmap =
			Files::instance()->kinematics()->equivalent_crossings();
	pair<INT, Crossing> uncrossed_i = Crossing::uncross(i);
	map<Crossing, Crossing>::const_iterator crossing;
	// minimal equivalent crossing
	crossing = crossmap.find(uncrossed_i.second);
	VERIFY(crossing != crossmap.end());
	// equivalent integral
	INT integral = SectorMappings::get_equivalent(uncrossed_i.first);

	const SectorMappings* m = Files::instance()->sectormappings(
			integral.integralfamily()->name());
	map<int, Sector>::const_iterator target = m->sector_relations().find(
			integral.id());
	map<int, exmap>::const_iterator shift = m->sector_relation_shifts().find(
			integral.id());

	// no shift available, no recursion
	if (target == m->sector_relations().end()) {
		if (integral != uncrossed_i.first || uncrossed_i.second
				!= crossing->second) {
			Identity outeq;
			outeq.insert(crossing->second.transform(integral), -1, false);
			outeq.remove_zero_integral();
			outeq.insert(i, 1, false);
			return outeq;
		} else {
			return Identity();
		}
	}
	VERIFY(shift != m->sector_relation_shifts().end());

	// shift available

	LinearCombination uncrossed = integral.apply_shift(shift->second,
			target->second.integralfamily(), -1, set_subsecs_to_zero);

	// recursion

	IdentityList replace;
	for (LinearCombination::const_iterator it = uncrossed.begin(); it
			!= uncrossed.end(); ++it)
		replace.push_back(get_shift_identity(it->first, set_subsecs_to_zero));

	for (IdentityList::const_iterator r = replace.begin(); r != replace.end(); ++r)
		uncrossed.replace(*r);

	LinearCombination crossed = crossing->second.transform(uncrossed);
	Identity outeq;
	outeq.insert(crossed.begin(), crossed.end());
	// lhs: the original integral
	outeq.insert(i, 1);
	return outeq;
}

void IdentityGenerator::genid_shift(const INT& integral, IdentityList& out,
		bool no_options) {
	IdentityList ident;
	ident.push_back(get_shift_identity(integral,
			options_.set_subsectors_to_zero_));
	if (!no_options)
		apply_remaining_options(ident);
	if (!ident.back().empty())
		out.splice(ident);
}

void IdentityGenerator::genid_shift(const set<INT>& in, IdentityList& out) {
	IdentityList identities;
	set<INT>::const_iterator it;
	for (it = in.begin(); it != in.end(); ++it)
		genid_shift(*it, identities, true);
	apply_remaining_options(identities);
	out.splice(identities);
}

void IdentityGenerator::genid_symmetry(const INT& integral, IdentityList& out,
		bool no_options) {
	IdentityList identities;
	int id = integral.id();
	const SectorMappings* m = mappings(integral.integralfamily());
	const map<int, list<exmap> >& all_shifts = m->sector_symmetries();
	map<int, list<exmap> >::const_iterator it = all_shifts.find(id);
	if (it == all_shifts.end())
		return;

	INT equiv = integral.get_equivalent();
	if (id != equiv.id()) {
		stringstream ss;
		ss << "Cannot use symmetry shifts for integral: ";
		integral.to_mma_stream(ss);
		ss << "\n since it reduces (by perm. sym.) to another sector: ";
		equiv.to_mma_stream(ss);
		ABORT(ss.str());
	}

	const list<exmap>& shifts = it->second;
	for (list<exmap>::const_iterator s = shifts.begin(); s != shifts.end(); ++s) {
		// rhs
		identities.push_back(
				integral.apply_shift(*s, integral.integralfamily(), -1,
						options_.set_subsectors_to_zero_));
		// lhs
		identities.back().insert(integral, 1);
	}

	if (!no_options)
		apply_remaining_options(identities);

	out.splice(identities);
}

const EquationList2& IdentityGenerator::get_template_equation(
		const IntegralFamily* ic, const string& type) {
	map<string, const IBPTemplate*>::const_iterator i;
	map<string, EquationList2>::const_iterator e;
	if ((i = template_by_ic_.find(ic->name())) != template_by_ic_.end()) {
		if ((e = i->second->template_eqs_by_keyword_.find(type))
				== i->second->template_eqs_by_keyword_.end())
			ABORT("Unknown type " << type);
		return e->second;
	}
	template_by_ic_[ic->name()] = new IBPTemplate(ic);
	return get_template_equation(ic, type);
}

const SectorMappings* IdentityGenerator::mappings(const IntegralFamily* ic) {
	map<string, const SectorMappings*>::const_iterator it;
	if ((it = mappings_by_ic_.find(ic->name())) != mappings_by_ic_.end()) {
		return it->second;
	} else {
		try {
			mappings_by_ic_[ic->name()] = Files::instance()->sectormappings(
					ic->name());
		} catch (exception& e) {
			ABORT("Cannot generate equations\n" << e.what());
		}
	}
	return mappings(ic);
}

void IdentityGenerator::apply_remaining_options(IdentityList& identities) const {
	if (options_.map_on_integral_)
		identities.transform_integrals(options_.map_on_integral_);
	if (!options_.map_on_coeff_.empty())
		identities.apply(options_.map_on_coeff_);
	identities.remove_zero_integral();
	if (options_.solve_equations_)
		identities.solve();
}

IdentityGenerator::IBPTemplate::IBPTemplate(const IntegralFamily* ic) :
	ic_(ic) {
	init_diff_ops();
	generate_ibp_template();
	generate_lorentz_template();
}

IdentityGenerator::IBPTemplate::~IBPTemplate() {
}

IdentityGenerator::IBPTemplate::IBPTemplate() :
	ic_(0) {
}

void IdentityGenerator::IBPTemplate::init_diff_ops() {
	// determine effect of q_k d/dq_i on general integrand,
	// expressed in terms of shift operators
	vector<int> v(2, 0);

	const lst loop_mom = ic_->loop_momenta();
	const lst iext_mom = ic_->kinematics()->independent_external_momenta();
	lst all_mom = add_lst(loop_mom, iext_mom);

	const vector<Propagator> propagators = ic_->propagators();
	const exmap s2i = ic_->kinematics()->rules_sp_to_invariants();
	const exmap s2p = ic_->rules_sp_to_prop();
	for (size_t i = 0; i < all_mom.nops(); i++) {
		for (size_t k = 0; k < all_mom.nops(); k++) {
			ASSERT(is_a<symbol>(all_mom[i]) && is_a<symbol>(all_mom[k]));
			const symbol qi = ex_to<symbol> (all_mom[i]);
			const symbol qk = ex_to<symbol> (all_mom[k]);
			v[0] = i;
			v[1] = k;
			diff_ops_[v] = OPSUM();
			for (size_t j = 0; j < propagators.size(); j++) {
				// qi (d/dqk)propj^aj = (aj propj^(aj-1)) * (qi dpropj/dqk)
				ex aj = ic_->propagator_exponents()[j];
				OP yjm = OP(-(j + 1)); // decrements aj by one
				ex qi_dpropj_dqk = propagators[j].derivative_contracted(qi, qk);
				qi_dpropj_dqk = qi_dpropj_dqk.expand();
				int opt = subs_options::algebraic;
				qi_dpropj_dqk = qi_dpropj_dqk.subs(s2p, opt);
				qi_dpropj_dqk = qi_dpropj_dqk.subs(s2i, opt);
				//LOG("qi_dpropj_dqk(" << i << "," << j << "," << k << ")="<<qi_dpropj_dqk);
				diff_ops_[v] += OPSUM(OPPROD(aj, yjm)) * ex_to_OPSUM(
						qi_dpropj_dqk, propagators);
			}
		}
	}
}

void IdentityGenerator::IBPTemplate::generate_ibp_template() {
	map<vector<int> , OPSUM> ibp_operators;

	// generate the ibp_
	EquationList2 ibp_, ibp_dim_, ibp_dim_free_, ibp_lee_;
	// (k,i)  ==>  d/dq_k * q_i
	vector<int> v(2, 0), dv(2, 0);
	size_t num_loop_mom = ic_->loop_momenta().nops();
	size_t num_iext_mom =
			ic_->kinematics()->independent_external_momenta().nops();
	// derivative:
	for (size_t k = 0; k < num_loop_mom; k++) {
		v[0] = k;
		dv[1] = k;
		// multiply:
		for (size_t i = 0; i < num_loop_mom + num_iext_mom; i++) {
			v[1] = i;
			dv[0] = i;
			Equation2 equibp;
			INT2 integral(ic_);
			ibp_operators[v] = diff_ops_[dv];
			ex dim = Files::instance()->kinematics()->dimension();
			if (i == k)
				ibp_operators[v] += OPSUM(OPPROD(dim, OP()));
			integral.operate(ibp_operators[v], equibp);
			ibp_.l.push_back(equibp);
			if (i != k)
				ibp_dim_free_.l.push_back(equibp);
			else
				ibp_dim_.l.push_back(equibp);
		}
	}
	template_eqs_by_keyword_["ibp"] = ibp_;
	template_eqs_by_keyword_["ibp_dim"] = ibp_dim_;
	template_eqs_by_keyword_["ibp_dim_free"] = ibp_dim_free_;

	// generate the ibp_lee_
	if (num_loop_mom >= 2) {
		//	vector<int> v(2, 0);
		v.assign(2, 0);

		// d/dk_i k_{i+1} , k_{l+1} == k_1, k_i = loop mom  ==> NoLM equations
		for (size_t i = 0; i < num_loop_mom; i++) {
			v[0] = i;
			v[1] = (i + 1) % num_loop_mom;
			INT2 integral(ic_);
			Equation2 equ;
			integral.operate(ibp_operators[v], equ);
			ibp_lee_.l.push_back(equ);
		}

		// d/dk_1 p_j , k_1 = loop mom, p_j ext. mom.   ==> NoEM equations
		for (size_t i = 0; i < num_iext_mom; i++) {
			v[0] = 1;
			v[1] = num_loop_mom + i;
			INT2 integral(ic_);
			Equation2 equ;
			integral.operate(ibp_operators[v], equ);
			ibp_lee_.l.push_back(equ);
		}

		// Sum_{i=0}^{NoLM-1} d/dk_i k_i , k_i = loop mom.   ==> 1 equation
		OPSUM opsum;
		for (size_t i = 0; i < num_loop_mom; i++) {
			v[0] = i;
			v[1] = i;
			opsum += ibp_operators[v];
		}
		INT2 integral(ic_);
		Equation2 equ;
		integral.operate(opsum, equ);
		ibp_lee_.l.push_back(equ);

		template_eqs_by_keyword_["ibp_lee"] = ibp_lee_;
	} else {
		//		cout << "The reduced set of identities (according to R.N.Lee) ";
		//		cout << "can only be used with more than 1 loop momenta" << endl;
	}
}

void IdentityGenerator::IBPTemplate::generate_lorentz_template() {
	EquationList2 lorentz_;
	lst loop_mom = ic_->loop_momenta();
	lst iext_mom = ic_->kinematics()->independent_external_momenta();
	const size_t num_loop_mom = loop_mom.nops();
	const size_t num_iext_mom = iext_mom.nops();
	lst all_mom = add_lst(loop_mom, iext_mom);
	//  Akj - Ajk; 0 <= k < NoEM,  k+1 <= j < NoEM
	// 	Akj = Sum_{i=0}^{NoEM-1}  ( p_j*p_i   p_k*d/dp_i   -   p_k*p_i   p_j*d/dp_i    )
	ex mom_i, mom_j, mom_k, pjpi, pkpi;
	OPSUM opsum1;
	OPSUM opsum2;
	vector<int> v1(2, 0), v2(2, 0);
	for (size_t k = num_loop_mom; k < num_iext_mom + num_loop_mom; k++) {
		v1[0] = k;
		for (size_t j = k + 1; j < num_iext_mom + num_loop_mom; j++) {
			v2[0] = j;
			OPSUM opsum;
			// calculate Akj and Ajk
			for (size_t i = num_loop_mom; i < num_iext_mom + num_loop_mom; i++) {
				mom_i = all_mom[i];
				mom_j = all_mom[j];
				mom_k = all_mom[k];
				exmap s2i = ic_->kinematics()->rules_sp_to_invariants();
				int opts = subs_options::algebraic;
				pkpi = ScalarProduct(mom_k, mom_i).eval().subs(s2i, opts);
				pjpi = ScalarProduct(mom_j, mom_i).eval().subs(s2i, opts);
				if (!freeof(pkpi, iext_mom) || !freeof(pjpi, iext_mom))
					ABORT("Failed to generate lorentz identities\n"
							<< pjpi << '\n' << pkpi << '\n' << s2i);
				v1[1] = i;
				v2[1] = i;
				opsum1 = diff_ops_[v1];
				opsum2 = diff_ops_[v2];
				opsum1 *= OPSUM(OPPROD(pjpi, OP()));
				opsum2 *= OPSUM(OPPROD(pkpi, OP()));
				opsum += opsum1 - opsum2;
			}
			INT2 integral(ic_);
			Equation2 equ;
			integral.operate(opsum, equ);
			lorentz_.l.push_back(equ);
		}
	}
	template_eqs_by_keyword_["lorentz"] = lorentz_;
}

// application of e to integrand is implied (e integrand = OPSUM integrand)
OPSUM ex_to_OPSUM(const GiNaC::ex& e, const vector<Propagator>& propagators) {
	exmap s;
	OPSUM res; // is zero !!!
	if (is_a<add> (e)) {
		for (size_t i = 0; i < e.nops(); ++i)
			res += ex_to_OPSUM(e.op(i), propagators);
	} else if (is_a<mul> (e)) {
		res = OPSUM(OPPROD(OP())); // is one
		for (size_t i = 0; i < e.nops(); ++i)
			res *= ex_to_OPSUM(e.op(i), propagators);
	} else if (e.has(Propagator(wild(1), wild(2))) || //
			e.has(Propagator(wild(1), wild(2), wild(3)))) {
		ex base, expo;
		if (e.match(pow(Propagator(wild(1), wild(2)), wild(3)), s)) {
			base = Propagator(wild(1).subs(s), wild(2).subs(s));
			expo = wild(3).subs(s);
		} else if (e.match(pow(Propagator(wild(1), wild(2), wild(3)), wild(4)),
				s)) {
			base = Propagator(wild(1).subs(s), wild(2).subs(s),//
					wild(3).subs(s));
			expo = wild(4).subs(s);
		} else if (e.match(Propagator(wild(1), wild(2)), s)) {
			base = Propagator(wild(1).subs(s), wild(2).subs(s));
			expo = 1;
		} else if (e.match(Propagator(wild(1), wild(2), wild(3)), s)) {
			base = Propagator(wild(1).subs(s), wild(2).subs(s),//
					wild(3).subs(s));
			expo = 1;
		} else {
			ABORT("Unknown structure containing propagator: " << e);
		}
		if (!(expo.info(info_flags::posint) || expo.info(info_flags::negint)))
			ABORT("Failed to parse propagator exponent in " << e);
		int n = ex_to<numeric> (expo).to_int();
		size_t i;
		for (i = 0; i < propagators.size(); ++i)
			if (base.match(propagators[i]))
				break;
		if (i == propagators.size())
			ABORT("Unknown propagator " << base);
		res = OPSUM(OPPROD(OP())); // is one
		for (size_t j = 0; j < static_cast<size_t> (abs(n)); ++j)
			res *= OPSUM(OP(n > 0 ? (i + 1) : -(i + 1)));
	} else if (!e.has(Propagator(wild(1), wild(2))) && //
			!e.has(Propagator(wild(1), wild(2), wild(3)))) {
		res = OPSUM(OPPROD(e, OP()));
	} else {
		ABORT("Failed to parse " << e);
	}
	return res;
}

LinearCombination IdentityGenerator::IBPTemplate::get_derivative_wrt_external_momenta(
		const INT& integral, int i, int k) {
	ASSERT(integral.integralfamily() == ic_);
	vector<int> v(2);
	v[0] = ic_->loop_momenta().nops() + i;
	v[1] = ic_->loop_momenta().nops() + k;
	INT2 genint(integral.integralfamily());
	Equation2 genlc;
	genint.operate(diff_ops_[v], genlc);
	return genlc.apply(integral);
}

} // namespace Reduze
