/*  job_setupsectormappings.cpp
 *
 *  Copyright (C) 2010-2012 Andreas von Manteuffel
 *  Copyright (C) 2010-2012 Cedric Studerus
 *
 *  This file is part of the package Reduze 2.
 *  It is distributed under the GNU General Public License version 3
 *  (see the file GPL-3.0.txt or http://www.gnu.org/licenses/gpl-3.0.txt).
 */

#include "job_setupsectormappings.h"
#include "functions.h"
#include "sectormappings.h"
#include "files.h"
#include "graph.h"

using namespace std;

namespace Reduze {

// register job type
namespace {
JobProxy<SetupSectorMappings> dummy;
}

SetupSectorMappings::SetupSectorMappings() :
	find_sector_relations_(true), find_sector_symmetries_(true),
			find_zero_sectors_(true), minimize_graphs_by_twists_(true),
			construct_minimal_graphs_(false), minimize_target_crossings_(true) {
	add_auto_options();
}

void SetupSectorMappings::run_serial() {

	// sort families by loop

	Files* files = Files::instance();
	map<int, list<IntegralFamily*> > families_by_loop;
	list<IntegralFamily*> ifs = files->integralfamilies();
	for (list<IntegralFamily*>::const_iterator f = ifs.begin(); f != ifs.end(); ++f)
		if (!(*f)->is_derived())
			families_by_loop[(*f)->loop_momenta().nops()].push_back(*f);

	// print what will be done

	map<int, list<IntegralFamily*> >::const_reverse_iterator fbl; // heavy work first
	LOG("Setup sector mappings for the following integral families requested:");
	for (fbl = families_by_loop.rbegin(); fbl != families_by_loop.rend(); ++fbl) {
		list<IntegralFamily*>::const_iterator fam;
		LOG("  " << fbl->first << "-loop:");
		for (fam = fbl->second.begin(); fam != fbl->second.end(); ++fam) {
			LOG("    " << (*fam)->name());
		}
	}
	LOG("");

	// setup the sector mappings for each class of families with the same number of loop momenta

	for (fbl = families_by_loop.rbegin(); fbl != families_by_loop.rend(); ++fbl) {
		// shift target sectors of all families with the same number of loop momenta
		set<SectorGL> shift_targets, dummy;
		bool lower_family_changed = false;
		list<IntegralFamily*>::const_iterator fam;
		for (fam = fbl->second.begin(); fam != fbl->second.end(); ++fam) {
			string sm = files->get_filename_sectormappings((*fam)->name());
			if (is_readable_file(sm) && is_conditional()) {
				// case of where a family is already setup: we must find the shift target sectors of this family
				SectorMappings* mp = files->sectormappings((*fam)->name());
				LOG("Integral family " << (*fam)->name() << " already setup.");
				// case where lower family has been changed not handled
				if (lower_family_changed)
					WARNING("Lower integral family has been changed."
							<< " This might invalidate the already existent setup"
							<< " of integral family " << (*fam)->name());
				if (find_sector_relations_)
					mp->find_shift_targets_with_graph(shift_targets,
							minimize_graphs_by_twists_,
							construct_minimal_graphs_);
			} else {
				// create empty sector mappings and set it up
				files->save(SectorMappings(*fam));
				SectorMappings* m = files->sectormappings((*fam)->name());
				m->setup(shift_targets, dummy, find_sector_relations_,
						find_sector_symmetries_, find_zero_sectors_,
						minimize_graphs_by_twists_, construct_minimal_graphs_,
						minimize_target_crossings_);
				LOG("Saving sector mappings for " << (*fam)->name());
				files->save(*m);
				LOG("");
				// LOGX("sector mappings for crossed integral families:");
				// for (list<IntegralFamily*>::const_iterator f = ifs.begin(); f
				// 		!= ifs.end(); ++f) {
				// 	if ((*f)->source_integralfamily()->name() == (*fam)->name()) {
				// 		YAML::Emitter ye;
				// 		ye << *(files->sectormappings((*f)->name()));
				// 		LOGX(ye.c_str());
				// 	}
				// }
			}
		} // t
	} // loop
}

bool SetupSectorMappings::find_dependencies(const set<string>& outothers,//
		list<string>& in, list<string>& out, list<Job*>& auxjobs) {
	Files* files = Files::instance();
	list<IntegralFamily*> ics = files->integralfamilies();
	list<IntegralFamily*>::const_iterator ic;
	for (ic = ics.begin(); ic != ics.end(); ++ic)
		if (!(*ic)->is_derived())
			out.push_back(files->get_filename_sectormappings((*ic)->name()));
	return true;
}

void SetupSectorMappings::init() {
}

std::string SetupSectorMappings::get_description() const {
	return string("setup sector mappings");
}

}
