/*  sectormappings.cpp
 *
 *  Copyright (C) 2010-2012 Andreas von Manteuffel
 *  Copyright (C) 2010-2012 Cedric Studerus
 *
 *  This file is part of the package Reduze 2.
 *  It is distributed under the GNU General Public License version 3
 *  (see the file GPL-3.0.txt or http://www.gnu.org/licenses/gpl-3.0.txt).
 */

#include "sectormappings.h"

#include "kinematics.h"
#include "int.h"
#include "equationlist.h"
#include "identitygenerator.h"
#include "functions.h"
#include "files.h"
#include "ginacutils.h"
#include "yamlutils.h"
#include "graph.h"

using namespace std;
using namespace GiNaC;

namespace Reduze {

// constructors

SectorMappings::SectorMappings(const IntegralFamily* ic) :
	integralfamily_(ic) {
}

SectorMappings::SectorMappings(const CrossedIntegralFamily* ic) :
	integralfamily_(ic) {
	string sn = ic->source_integralfamily()->name();
	const SectorMappings* sm = Files::instance()->sectormappings(sn);
	zero_sectors_ = sm->zero_sectors_;
	sectors_without_graph_ = sm->sectors_without_graph_;
}

void SectorMappings::setup(std::set<SectorGL>& shift_targets, //
		std::set<SectorGL>& local_shift_targets, //
		bool find_sector_relations, //
		bool find_sector_symmetries, //
		bool find_zero_sectors, //
		bool minimize_by_twists, //
		bool construct_minimal_graphs, //
		bool minimize_target_crossings) {

	local_shift_targets.clear();

	const Files* files = Files::instance();
	const IntegralFamily* ic = integralfamily_;
	const int max_sec_id = ic->max_sector_id();
	const bool allow_long_zero_test =
			ic->kinematics()->external_momenta().nops() != 0;
	LOG("Setup sector mappings for " << ic->name());

	const list<Crossing>& oc = ic->kinematics()->ordered_crossings();
	LOGX("Ordered crossings:");
	for (list<Crossing>::const_iterator c = oc.begin(); c != oc.end(); ++c) {
		YAML::Emitter ye;
		ye << *c;
		LOGX("  " << ye.c_str());
	}
	LOGX("");

	YAML::Emitter e;
	ic->print(e);
	LOGX(e.c_str());
	LOGX("");
	LOGX("Rules scalar products to invariants:\n" << ic->kinematics()->rules_sp_to_invariants());
	LOGX("Rules invariants to scalar products:\n" << ic->kinematics()->find_rules_invariants_to_sp());
	LOGX("Rules scalar products to propagators:\n" << ic->rules_sp_to_prop());
	LOGX("");

	LOG("Maximal sector id is " << max_sec_id << ".");
	LOG(max_sec_id + 1 << " sectors occur in total");

	// sector tree (using global permutation symmetry)

	LOG("Finding non-equivalent sectors using permutation symmetries:");
	bool perm_sym = !ic->cycles().empty();
	set<Sector> sectors;
	set<int> sector_ids;
	for (int i = 0; i <= max_sec_id; ++i) {
		int id = perm_sym ? ic->get_equivalent(i) : i;
		sectors.insert(Sector(ic, id));
		sector_ids.insert(sector_ids.end(), id);
	}
	// print non-equivalent sectors to log-file
	LOGX("\nNon-equivalent sectors:");
	LOGX(ic->t_tree_nice_string(sector_ids));
	LOG(sector_ids.size() << " non-equivalent sectors were found");

	// graphs of sectors
	// sectors with vanishing jacobian determinant of the loop momenta matrix are added to zero_sectors_

	map<Sector, Graph> graph_by_sector;
	find_graphs(graph_by_sector, sectors, minimize_by_twists,
			construct_minimal_graphs);
	//find_graphs_from_top(graph_by_sector, sectors, minimize_by_twists);
	LOGX("\nNo graph sectors:");
	LOGX(ic->t_tree_nice_string(sectors_without_graph()));

	// canonical labeling

	LOG("Finding canonical labels:");
	ProgressBar pbar(2, "canonical labels:", graph_by_sector.size());
	pbar.start();
	set<SectorGL> local_sectors;
	for (map<Sector, Graph>::const_iterator s = graph_by_sector.begin(); s
			!= graph_by_sector.end(); ++s) {
		local_sectors.insert(SectorGL(s->first, s->second,
				s->second.find_canonical_graph_label()));
		pbar.print();
	}
	pbar.end();

	// list of sectors by their shift target (all sectors are from this family, target not included in the set)
	map<SectorGL, list<SectorGL> > local_targets_by_equiv;

	// try to shift the 'local_sectors' to the shift_targets or to them self
	// updates the 'shift_targets'
	// sets the 'local_shift_targets'
	if (find_sector_relations) {
		find_sector_shifts(shift_targets, local_sectors,
				local_targets_by_equiv, minimize_target_crossings);
	} else {
		// move local sectors to local_shift_targets
		map<SectorGL, list<SectorGL> >::iterator in =
				local_targets_by_equiv.end();
		list<SectorGL> empty;
		for (set<SectorGL>::iterator s = local_sectors.begin(); s
				!= local_sectors.end();) {
			in = local_targets_by_equiv.insert(in, make_pair(*s, empty));
			local_sectors.erase(s++);
		}
	}

	// zero test for the local_targets
	if (find_zero_sectors) {
		LOGN("Finding zero sectors of sectors with a graph:");
		map<Sector, list<Sector> > zeros;
		map<Sector, list<Sector> >::iterator in = zeros.end();
		map<SectorGL, list<SectorGL> >::const_iterator s;
		list<SectorGL>::const_iterator t;
		for (s = local_targets_by_equiv.begin(); s
				!= local_targets_by_equiv.end(); ++s) {
			in = zeros.insert(in, make_pair(s->first.sector_, list<Sector> ()));
			for (t = s->second.begin(); t != s->second.end(); ++t)
				in->second.push_back(t->sector_);
		}
		this->find_zero_sectors(zeros, allow_long_zero_test);
		LOG("Found " << zero_sectors_.size() << " zero sectors (isomorphics included).");
	}

	set<int> non_zero_shift_targets;
	set<GraphLabel> local_target_labels;
	for (map<SectorGL, list<SectorGL> >::const_iterator s =
			local_targets_by_equiv.begin(); s != local_targets_by_equiv.end(); ++s) {
		if (!is_in_zero_sectors_non_static(s->first.sector_.id())) {
			non_zero_shift_targets.insert(s->first.sector_.id());
			local_target_labels.insert(s->first.labels_.first);
		}
		local_shift_targets.insert(s->first);
	}
	LOGX("\nNon-vanishing shift targets:");
	LOGX(ic->t_tree_nice_string(non_zero_shift_targets));
	LOG("Found " << non_zero_shift_targets.size() << " non-vanishing shift target sectors");
	// the non-vanishing local shift targets can still contain graphs which are isomorphic
	// but a shift between them would have |det| != 1. Log info:
	int num_same_graph = non_zero_shift_targets.size()
			- local_target_labels.size();
	LOG("  " << local_target_labels.size() << " of them "
			<< (num_same_graph == 1 ? "has" : "have")
			<< " different graphs.");

	// symmetry shifts

	if (find_sector_symmetries) {
		LOG("Finding symmetry shift relations:");
		SectorMappings::find_symmetry_shifts(local_shift_targets);
	}
	//LOG("");

	// print graphs as dot files in standard directory ./graphs/family_name/

	const string graph_dir = files->get_graph_directory(ic);
	LOG("Printing graphs from sectors in " << graph_dir << ":");
	remove_directory_with_files(graph_dir);
	make_directory(graph_dir);
	const string all_dir = graph_dir + "all/";
	const string shift_targets_dir = graph_dir + "shift_targets/";
	make_directory(all_dir);
	make_directory(shift_targets_dir);

	string dot_file_prefix = all_dir + safe_filename(ic->name());
	string dot_file_prefix2 = shift_targets_dir + safe_filename(ic->name());
	string ps_file = safe_filename(ic->name()) + ".ps";
	string ps_file2 = safe_filename(ic->name()) + "_shift_targets.ps";
	for (map<Sector, Graph>::const_iterator s = graph_by_sector.begin(); s
			!= graph_by_sector.end(); ++s)
		s->second.print_dot(all_dir + s->first.get_safe_string_for_filename(
				true) + ".dot");
	for (set<SectorGL>::const_iterator s = local_shift_targets.begin(); s
			!= local_shift_targets.end(); ++s)
		//if (!is_zero_sector_non_static(s->sector_.id()))
		s->graph_.print_dot(shift_targets_dir
				+ s->sector_.get_safe_string_for_filename(true) + ".dot");

	LOG("To generate a postscript file from the graphs in the dot format type e.g.");
	LOG("  dot -Tps -o " << ps_file << " " << dot_file_prefix << "_*.dot");
	LOG("  dot -Tps -o " << ps_file2 << " " << dot_file_prefix2 << "_*.dot");
	//LOG("");
	LOG("Done setup sector mappings for " << ic->name());
}

bool SectorMappings::is_in_sectors_without_graph(const Sector& sector) {
	const SectorMappings* m = Files::instance()->sectormappings(
			sector.integralfamily()->name());
	const set<int>& no_graphs = m->sectors_without_graph();
	return no_graphs.find(sector.id()) != no_graphs.end();
}
bool SectorMappings::is_in_sectors_without_graph_nonstatic(int sector_id) const {
	return sectors_without_graph_.find(sector_id)
			!= sectors_without_graph_.end();
}

bool SectorMappings::is_in_zero_sectors(const Sector& sector) {
	const SectorMappings* m = Files::instance()->sectormappings(
			sector.integralfamily()->name());
	const set<int>& zeros = m->zero_sectors();
	return zeros.find(sector.id()) != zeros.end();
}

bool SectorMappings::is_in_zero_sectors_non_static(int sector_id) const {
	return zero_sectors_.find(sector_id) != zero_sectors_.end();
}

// get equivalent and reduced sectors

// static
INT SectorMappings::get_equivalent(const INT& i) {
	INT red = i.get_equivalent();
	const IntegralFamily* ic_red = red.integralfamily();
	const SectorMappings* m = Files::instance()->sectormappings(ic_red->name());
	if (m->zero_sectors_.find(red.id()) != m->zero_sectors_.end())
		return INT(ic_red); // zero integral for #loops>0, irrelevant for tree
	return red;
}

// non-static
int SectorMappings::get_equivalent(int sector_id) const {
	INT corner(integralfamily_, sector_id);
	return get_equivalent_nonstatic(corner).id();
}

// static
Sector SectorMappings::get_equivalent(const Sector& s) {
	string name = s.integralfamily()->name();
	SectorMappings* m = Files::instance()->sectormappings(name);
	return Sector(m->integralfamily_, m->get_equivalent(s.id()));
}

// static
Sector SectorMappings::get_shifted_of_equivalent(const Sector& s) {
	string name = s.integralfamily()->name();
	SectorMappings* m = Files::instance()->sectormappings(name);
	return m->get_shifted_of_equivalent(s.id());
}

// non-static
Sector SectorMappings::get_shifted_of_equivalent(int from_id) const {
	int equiv_id = get_equivalent(from_id);
	map<int, Sector>::const_iterator it = sector_relations_.find(equiv_id);
	if (it == sector_relations_.end())
		return Sector(integralfamily_, equiv_id);
	return it->second;
}

// static
Sector SectorMappings::get_equivalent_of_shifted_of_equivalent(const Sector& s) {
	Sector ns = SectorMappings::get_shifted_of_equivalent(s);
	ns = SectorMappings::get_equivalent(ns);
	return ns;
}

// static
Sector SectorMappings::get_reduced_sector(const Sector& s) {
	Sector ns = s;
	Sector os = s;
	do {
		os = ns;
		ns = SectorMappings::get_equivalent_of_shifted_of_equivalent(ns);
	} while (ns != os);
	return ns;
}

std::set<Sector> SectorMappings::get_reduced_subsectors(const Sector& s,
		bool recurse) {
	set<Sector> res;
	stack<Sector> todo;
	Sector next = SectorMappings::get_reduced_sector(s);
	do { // find reduced immediate sub sectors of next
		const IntegralFamily* f = next.integralfamily();
		set<int> subs = f->get_subsector_equivalents(next.id(), false);
		for (set<int>::const_iterator s = subs.begin(); s != subs.end(); ++s) {
			Sector sr = SectorMappings::get_reduced_sector(Sector(f, *s));
			if (!sr.is_zero() && res.find(sr) == res.end()) {
				res.insert(sr);
				todo.push(sr);
			}
		}
	} while (recurse && !todo.empty() && (next = todo.top(), todo.pop(), true));
	return res;
}

// \todo remove me
INT SectorMappings::get_equivalent_nonstatic(INT i) const {
	ASSERT(*(i.integralfamily()) == *integralfamily_);
	INT red = i.get_equivalent();
	if (zero_sectors_.find(red.id()) != zero_sectors_.end())
		return INT(integralfamily_); // zero integral
	return red;
}

bool SectorMappings::search_corner_reduction_to_zero(int sector_id,
		bool allow_longtest) const {
	Sector sector(integralfamily_, sector_id);
	LOGXX("Checking if sector " << sector << " is zero");
	// for tree level integral family, t=0 is not zero
	if (sector.is_zero())
		return LOGXX("Is trivial zero sector"), true;
	if (sector.is_non_zero())
		return LOGXX("Is trivial tree level (non-zero) sector"), false;
	if (zero_sectors_.find(sector_id) != zero_sectors_.end())
		return LOGXX("Is already known to be zero"), true;
	INT corner(integralfamily_, sector_id);
	int t = corner.t();
	INT (SectorMappings::*trafo)(INT) const =//
			&SectorMappings::get_equivalent_nonstatic;
	set<INT> seeds;
	seeds.insert(corner);
	IdentityList idlist;
	exmap s2o;
	const symbol* sym =
			integralfamily_->kinematics()->symbol_to_replace_by_one();
	if (sym != 0)
		s2o[*sym] = 1;
	IdentityGeneratorOptions options;
	options.set_map_on_coeff(s2o);
	options.set_eliminate_subsectors_by_shift(false); // otherwise throws, since sectormappings aren't known yet
	options.set_transform_integrals_to_equivalent(false); // otherwise throws, since sectormappings aren't known yet
	static IdentityGenerator generator(options); // static gives great speedup here
	//	generator.set_map_on_integral(trafo);
	generator.generate(seeds, idlist, "ibp");
	idlist.transform_integrals(bind1st(mem_fun(trafo), this));
	idlist.remove_zero_integral();
	idlist.remove_empty_equations();
	if (idlist.size() == 0)
		WARNING("Empty identity list in zero test for sector " << sector_id);
	LOGXX(idlist);
	idlist.reduze();
	if (idlist.gives_reduction_zero(corner))
		return LOGXX("Is zero since corner reduces to zero"), true;
	// increasing s2 can't help for maximal sector
	if (sector_id == integralfamily_->max_sector_id())
		return LOGXX("Treat maximal sector as non-zero"), false;
	// test with more equations, s2 -> s2 + 1
	if (allow_longtest) {
		SeedGeneratorOptions seed_options;
		seed_options.set_reduce_to_equivalent(true);
		SeedGenerator seed_generator(integralfamily_, seed_options);
		set<INT> seeds2;
		seed_generator.generate(sector_id, t, t, 1, 1, seeds2);
		IdentityList idlist2;
		generator.generate(seeds2, idlist2, "ibp");
		idlist2.splice(idlist);
		idlist2.transform_integrals(bind1st(mem_fun(trafo), this));
		idlist2.remove_zero_integral();
		idlist2.remove_empty_equations();
		idlist2.reduze();
		if (idlist2.gives_reduction_zero(
				set<INT> (seeds2.begin(), seeds2.end())))
			return LOGXX("Is zero since extended seeds reduce to zero"), true;
	}
	return LOGXX("Not found to reduce to zero"), false;
}

void SectorMappings::get_graphs(std::map<Sector, Graph>& graph_by_sector,
		const std::set<Sector>& sectors, bool minimize_by_twists,
		bool construct_minimal_graphs) const {
	string twistonoff = (minimize_by_twists ? "on" : "off");
	string minonoff = (construct_minimal_graphs ? "on" : "off");
	ProgressBar pbar(2, "finding graphs: (twists " + twistonoff
			+ " / construct minimal " + minonoff + ")", sectors.size());
	pbar.start();
	pair<map<Sector, Graph>::iterator, bool> inserted;
	for (set<Sector>::const_iterator s = sectors.begin(); s != sectors.end(); ++s) {
		pbar.print();
		bool det_non_zero;
		Graph g;
		if (s->search_graph(g, det_non_zero, construct_minimal_graphs)) {
			if (minimize_by_twists)
				g.minimize_by_twists();
			inserted = graph_by_sector.insert(make_pair(*s, Graph()));
			ASSERT(inserted.second);
			inserted.first->second.swap(g);
		}
	}
	pbar.end();

}

void SectorMappings::find_graphs(std::map<Sector, Graph>& graph_by_sector,
		const std::set<Sector>& sectors, bool minimize_by_twists,
		bool construct_minimal_graphs) {
	LOG("Finding graphs of sectors (twists " << (minimize_by_twists ? "on" : "off")
			<< " / construct minimal " << (construct_minimal_graphs ? "on" : "off")
			<< "):");
	ProgressBar pbar(2, "finding graphs", sectors.size());
	pbar.start();
	pair<map<Sector, Graph>::iterator, bool> inserted;
	int t = -1, col = 1, num_zeros = 0;
	for (set<Sector>::const_iterator s = sectors.begin(); s != sectors.end(); ++s) {
		pbar.print();
		int id = s->id();
		if (s->t() != t) {
			t = s->t();
			col = 1;
			//LOGN(endl << "   t = " << t << ":\t\b\b\b\b" << flush);
		} // else if ((col++) % 60 == 0) {
		//	LOGN (endl << "\t\t\b\b\b\b" << flush);
		//}
		set<int> sub = integralfamily_->get_subsector_equivalents(id, false);
		set<int>::iterator n = sub.begin();
		for (; n != sub.end(); ++n)
			if (sectors_without_graph_.find(*n) != sectors_without_graph_.end())
				break;
		bool det_non_zero = true;
		Graph g;
		if (n != sub.end() || !(s->search_graph(g, det_non_zero,
				construct_minimal_graphs))) {
			if (det_non_zero) {
				sectors_without_graph_.insert(id);
				//LOGN("n" << flush);
			} else {
				++num_zeros;
				zero_sectors_.insert(id); // det = 0
				//LOGN("0" << flush);
			}
		} else {
			if (minimize_by_twists)
				g.minimize_by_twists();
			inserted = graph_by_sector.insert(make_pair(*s, Graph()));
			ASSERT(inserted.second);
			inserted.first->second.swap(g);
			//LOGN("g" << flush);
		}
	}
	pbar.end();
	LOG("Found " << graph_by_sector.size() << " graphs");
}

// Baustelle
//void SectorMappings::find_graphs_from_top(
//		std::map<Sector, Graph>& graph_by_sector, const std::set<Sector>& secs,
//		bool minimize_by_twists) {
//
//	LOG("Finding graphs of sectors (twists " << (minimize_by_twists ? "on" : "off") << "):");
//	set<Sector, greater<Sector> > sectors(secs.begin(), secs.end()); // copy
//	ProgressBar pbar(2, "finding graphs", sectors.size());
//	pbar.start();
//	pair<map<Sector, Graph>::iterator, bool> inserted;
//	int num_zeros = 0;
//	for (set<Sector, greater<Sector> >::iterator s = sectors.begin(); s
//			!= sectors.end(); ++s) {
//		VERIFY(*integralfamily_ == *s->integralfamily());
//		cerr << "find graph for sector " << *s << endl;
//		pbar.print();
//		int id = s->id();
//		bool det_non_zero = true;
//		Graph g;
//		if (!(s->search_graph(g, det_non_zero))) {
//			if (det_non_zero) {
//				sectors_without_graph_.insert(id);
//			} else {
//				++num_zeros;
//				zero_sectors_.insert(id); // det = 0
//			}
//		} else {
//			if (minimize_by_twists)
//				g.minimize_by_twists();
//			inserted = graph_by_sector.insert(make_pair(*s, Graph()));
//			ASSERT(inserted.second);
//			inserted.first->second.swap(g);
//			set<int> subs = integralfamily_->get_subsector_equivalents(
//					s->id(), true);
//			for (set<Sector, greater<Sector> >::iterator ss = s; ss
//					!= sectors.end();)
//				if (subs.find(ss->id()) != subs.end())
//					sectors.erase(ss++);
//				else
//					++ss;
//		}
//	}
//	pbar.end();
//	LOG("Found " << graph_by_sector.size() << " graphs");
//}

void SectorMappings::find_canonical_graph_labels(std::map<Sector, std::pair<
		GraphLabel, GraphRelabeling> >& labels, //
		const std::map<Sector, Graph>& sectors) {
	LOG("  Find canonical labels:");
	ProgressBar pbar(2, "canonical labels:", sectors.size());
	pbar.start();
	map<Sector, Graph>::const_iterator s;
	map<Sector, pair<GraphLabel, GraphRelabeling> >::iterator hint =
			labels.begin();
	for (s = sectors.begin(); s != sectors.end(); ++s) {
		hint = labels.insert(hint, make_pair(s->first,
				s->second.find_canonical_graph_label()));
		pbar.print();
	}
	pbar.end();
}

void SectorMappings::find_sector_shifts(
		std::set<SectorGL>& minimal, //
		const std::set<SectorGL>& sectors, //
		std::map<SectorGL, std::list<SectorGL> >& new_targets_with_equiv,
		bool minimize_target_crossings) {
	LOG("Finding sector shifts:");
	ProgressBar pbar(2, "sector shifts:", sectors.size());
	pbar.start();
	LOGX("");
	int num_shifts = 0;
	map<Sector, list<map<int, int> > > symmetry_shifts;
	for (set<SectorGL>::const_iterator s = sectors.begin(); s != sectors.end(); ++s) {
		pbar.print();
		const SectorGL& sec = *s;
		set<SectorGL>::const_iterator m;
		for (m = minimal.begin(); m != minimal.end(); ++m) {
			const SectorGL& min = *m;
			const IntegralFamily* f = min.sector_.integralfamily();
			ASSERT(*f == *f->source_integralfamily()); // min not from crossed integral family
			ASSERT(min.sector_ < sec.sector_); // min lower than sec
			ASSERT(*sec.sector_.integralfamily() == *integralfamily_); // sec belong to this integral family
			Crossing crossing(f->kinematics());
			GiNaC::exmap shift;
			GiNaC::ex det(0);
			if (min.labels_.first == sec.labels_.first) {
				det = Graph::find_shift(sec.graph_, sec.labels_, min.graph_,
						min.labels_, shift, crossing,
						symmetry_shifts[min.sector_]);
				Crossing inverse_crossing = crossing.inverse();
				string name = inverse_crossing.name_for_crossed_family(f);
				const Sector target = Sector(Files::instance()->integralfamily(
						name), min.sector_.id());
				const Sector& source = sec.sector_;
				if (det.is_equal(1)) {
					++num_shifts;
					LOGX("Found shift from " << source << " to " << target);
					insert_shift_nonstatic(source.id(), target, shift);
					// add uncrossed to new targets if it is not shifted to another family
					if (*(min.sector_.integralfamily()) == *integralfamily_)
						new_targets_with_equiv[min].push_back(sec);
					break;
				} else {
					LOGX("Reject shift with |det| = " << det << " from sector "
							<< source << " to " << target);
					LOGX("  shift: " << shift);
				}
			}
		}
		if (m == minimal.end()) { // sec not shifted to any sector in minimal
			LOGX("No shift found for sector " << sec.sector_);
			minimal.insert(minimal.end(), sec);
			if (minimize_target_crossings && sec.graph_.num_external_nodes()
					> 0)
				symmetry_shifts.insert(symmetry_shifts.end(), make_pair(
						sec.sector_, sec.graph_.find_node_symmetry_group()));
			new_targets_with_equiv[sec];
		}
	}
	pbar.end();
	LOG("  Found " << num_shifts << " new shifts to lower sectors");
}

void SectorMappings::insert_zero_sectors(const std::set<int>& sector_ids) {
	zero_sectors_.insert(sector_ids.begin(), sector_ids.end());
}

int SectorMappings::find_zero_sectors(
		const std::map<Sector, std::list<Sector> >& sectors_by_rep,
		bool allow_longtest) {
	int tt = -1, col = 1, num_zeros = 0;
	map<Sector, list<Sector> >::const_iterator m;
	for (m = sectors_by_rep.begin(); m != sectors_by_rep.end(); ++m) {
		ASSERT(*(m->first.integralfamily()) == *integralfamily_);
		const Sector& s = m->first;
		int id = s.id();
		if (s.t() != tt) {
			tt = s.t();
			col = 1;
			LOGN(endl << "   t = " << tt << ":\t\b\b\b\b" << flush);
		} else if ((col++) % 60 == 0) {
			LOGN (endl << "\t\t\b\b\b\b" << flush);
		}
		if (is_in_zero_sectors_non_static(id))
			continue;
		set<int> sub = integralfamily_->get_subsector_equivalents(id, false);
		for (set<int>::iterator n = sub.begin(); n != sub.end();)
			if (is_in_zero_sectors_non_static(*n)) {
				sub.erase(n++);
			} else if (Sector(integralfamily(), *n).is_zero()) {
				zero_sectors_.insert(*n);
				sub.erase(n++);
			} else {
				break; // non-zero sub sectors for id
			}
		// calculate if zero sector (skip test if non-zero subsectors present)
		if (sub.empty() && search_corner_reduction_to_zero(id, allow_longtest)) {
			zero_sectors_.insert(id); // is zero
			// include isomorphic zero sectors
			for (list<Sector>::const_iterator z = m->second.begin(); z
					!= m->second.end(); ++z) {
				ASSERT(*(z->integralfamily()) == *integralfamily_);
				zero_sectors_.insert(z->id());
			}
			LOGN("0" << flush);
			++num_zeros;
		} else {
			LOGN("x" << flush);
		}
	}
	LOG("");
	return num_zeros;
}

void SectorMappings::find_symmetry_shifts(const std::set<SectorGL>& sectors) {
	ProgressBar pbar(2, "finding symmetry shifts", sectors.size());
	pbar.start();
	for (set<SectorGL>::const_iterator s = sectors.begin(); s != sectors.end(); ++s) {
		ASSERT(*s->sector_.integralfamily() == *integralfamily_);
		list<exmap> shifts;
		s->graph_.find_symmetry_shifts(shifts,
				s->sector_.integralfamily()->kinematics());
		if (!shifts.empty())
			insert_symmetry_shifts_nonstatic(s->sector_.id(), shifts);
		pbar.print();
	}
	pbar.end();
}

//GiNaC::exmap SectorMappings::find_shift_for_sector_relation(INT& integral) const {
//	using namespace GiNaC;
//	if (*integral.integralfamily() != *integralfamily_)
//		throw runtime_error("can't find shift for alien integralfamily");
//	// permute to equivalent from sector to find shift
//	const map<int, exmap>& shifts = sector_relation_shifts_;
//	const map<int, vector<vector<int> > >& cyclemap =
//			integralfamily_->get_cycles_();
//	INT pint(integral);
//	map<int, exmap>::const_iterator trafo;
//	map<int, vector<vector<int> > >::const_iterator c = cyclemap.begin();
//	while ((trafo = shifts.find(pint.id())) == shifts.end()//
//			&& c != cyclemap.end()) {
//		pint = integral;
//		pint.transform(c->second);
//		++c;
//	}
//	if (trafo == shifts.end()) {
//		stringstream ss;
//		ss << "no shift found for sector " << integral.get_sector();
//		throw runtime_error(ss.str());
//	}
//	integral = pint;
//	return trafo->second;
//}

bool SectorMappings::test_shift(const Sector& from, const Sector& to,
		const GiNaC::exmap& shift) {
	INT integral(from.integralfamily(), from.id());
	Identity lhs;
	try {
		lhs = integral.apply_shift(shift, to.integralfamily(), 1, true);
	} catch (exception& e) {
		stringstream ss;
		ss << "shift " << shift << " doesn't map sector " << from << " to "
				<< to << "\n";
		vector<Propagator> p1 = from.get_propagators();
		vector<Propagator> p2 = to.get_propagators();
		ss << from << ": ";
		copy(p1.begin(), p1.end(), ostream_iterator<Propagator> (ss, " "));
		ss << "\n" << to << ": ";
		copy(p2.begin(), p2.end(), ostream_iterator<Propagator> (ss, " "));
		ss << "\n";
		LOG(ss.str() + e.what());
		return false;
	}
	if (lhs.size() == 1 && lhs.begin()->first.id() == to.id())
		return true;
	return false;
}

void SectorMappings::insert_shift_nonstatic(int from_id, const Sector& to,
		const GiNaC::exmap& shift) {

	const Sector from(integralfamily_, from_id);

	// from-sectors must be minimal w.r.t. global permutation symmetries.
	int from_equiv = integralfamily_->get_equivalent(from_id);
	if (from_equiv != from_id) {
		Sector min(integralfamily_, from_equiv);
		ABORT("Reject shift from " << from << " to sector " << to << ".\n"
				<< "Sector " << from << " is equivalent to sector " << min
				<< " by global permutation symmetries.\n"
				<< "Try a shift from " << min << " to " << to << " instead.");
	}

	// target sectors must be lower sectors than "from"-sectors.
	if (!(to < from))
		ABORT("Reject shift from lower (or equal) sector " << from << " to " << to << ".\n"
				<< "Try a shift from " << to << " to " << from);

	// shifts between physical and unphysical sectors are not allowed.
	bool same_family = (*integralfamily_ == *(to.integralfamily()));
	bool graph1 = !is_in_sectors_without_graph_nonstatic(from_id);
	bool graph2 =
			(same_family ? !is_in_sectors_without_graph_nonstatic(to.id())
					: !is_in_sectors_without_graph(to));
	if (graph1 != graph2) {
		if (graph1)
			ABORT("Reject shift between physical sector " << from << " and unphysical sector " << to);
		else
			ABORT("Reject shift between unphysical sector " << from << " and physical sector " << to);
	}

	if (!SectorMappings::test_shift(from, to, shift))
		ABORT("Wrong shift from sector " << from << " to sector "
				<< to << " detected: " << shift);

	pair<map<int, Sector>::const_iterator, bool> inserted1;
	pair<map<int, GiNaC::exmap>::const_iterator, bool> inserted2;
	inserted1 = sector_relations_.insert(make_pair(from_id, to));
	inserted2 = sector_relation_shifts_.insert(make_pair(from_id, shift));
	ASSERT(inserted1.second == inserted2.second);
	if (!inserted1.second)
		ABORT("Shift from " << from << " to " << to << " already inserted.");
}

void SectorMappings::insert_shift(const Sector& from, const Sector& to,
		const GiNaC::exmap& shift) {
	const string& name = from.integralfamily()->name();
	SectorMappings *m = Files::instance()->sectormappings(name);
	m->insert_shift_nonstatic(from.id(), to, shift);
}

void SectorMappings::insert_symmetry_shifts_nonstatic(int sector_id,
		const std::list<GiNaC::exmap>& shifts) {
	list<exmap>::const_iterator symm;
	Sector sec(integralfamily_, sector_id);
	for (symm = shifts.begin(); symm != shifts.end(); ++symm)
		if (!SectorMappings::test_shift(sec, sec, *symm))
			ABORT("Wrong symmetry shift for " << sec << " detected: " << *symm);
	pair<map<int, list<GiNaC::exmap> >::const_iterator, bool> inserted;
	inserted = sector_symmetries_.insert(make_pair(sector_id, shifts));
	if (!inserted.second)
		ABORT("Symmetry shifts for sector " << sec << " already inserted.");
}

void SectorMappings::insert_symmetry_shifts(const Sector& sec,
		const list<exmap>& shifts) {
	const string& name = sec.integralfamily()->name();
	SectorMappings *m = Files::instance()->sectormappings(name);
	m->insert_symmetry_shifts_nonstatic(sec.id(), shifts);
}

void SectorMappings::find_shift_targets(set<Sector>& sectors) const {
	int maxid = integralfamily_->max_sector_id();
	set<int> equiv = integralfamily_->get_subsector_equivalents(maxid, true);
	equiv.insert(maxid);
	// exclude sectors for which a shift exists
	for (set<int>::iterator e = equiv.begin(); e != equiv.end(); ++e)
		if (sector_relations_.find(*e) == sector_relations_.end()) // no shift found
			sectors.insert(Sector(integralfamily_, *e));
}

void SectorMappings::find_shift_targets_with_graph(
		std::set<SectorGL>& shift_targets, bool minimize_by_twists,
		bool construct_minimal_graphs) const {
	set<Sector> sectors;
	// minimize sector ids by permutation symmetries
	ProgressBar pbar(2, "finding shift targets:", 0);
	pbar.start();
	int maxid = integralfamily_->max_sector_id();
	set<int> equiv = integralfamily_->get_subsector_equivalents(maxid, true);
	equiv.insert(maxid);
	// exclude no graph sectors
	int num_equiv = equiv.size();
	pbar.set_max(num_equiv);
	for (set<int>::iterator e = equiv.begin(); e != equiv.end();) {
		pbar.print();
		if (sectors_without_graph_.find(*e) != sectors_without_graph_.end())
			equiv.erase(e++);
		else
			++e;
	}
	pbar.set_max(num_equiv + equiv.size());
	// exclude sectors for which a shift exists
	for (set<int>::iterator e = equiv.begin(); e != equiv.end(); ++e) {
		pbar.print();
		if (sector_relations_.find(*e) == sector_relations_.end()) // no shift found
			sectors.insert(Sector(integralfamily_, *e));
	}
	pbar.end();

	map<Sector, Graph> graph_by_sec;
	get_graphs(graph_by_sec, sectors, minimize_by_twists,
			construct_minimal_graphs);
	pbar.set_max(graph_by_sec.size());
	pbar.set_title("canonical labels:");
	pbar.start();
	for (map<Sector, Graph>::const_iterator s = graph_by_sec.begin(); s
			!= graph_by_sec.end(); ++s) {
		shift_targets.insert(SectorGL(s->first, s->second,
				s->second.find_canonical_graph_label()));
		pbar.print();
	}
	pbar.end();
}

/// \todo: idea for improvement:
// show separate t graph: 4->3->2->1->0 vertically
// and show sectors at appropriate vert. position
// similarly to timeline example on http://www.graphviz.org/pdf/dotguide.pdf
// page 16/17
void SectorMappings::print_sectortree_dot(const std::set<Sector>& secs,
		const std::string& filename) {
	string filename_tmp = filename + ".tmp";
	ofstream file(filename_tmp.c_str());

	// header
	file << "digraph SectorRelations {\n";
	file << "size = \"8.3,11.7\"\n";
	file << "ratio = \"1.5\"\n";
	file << "center = \"true\"\n";

	set<Sector> allsecs = secs;
	set<Sector>::const_iterator s;
	for (s = secs.begin(); s != secs.end(); ++s) {
		Sector equiv(SectorMappings::get_equivalent(*s));
		if (equiv == *s)
			continue;
		allsecs.insert(equiv);
		file//
				<< s->integralfamily()->name()//
				<< s->id()//
				<< " -> "//
				<< equiv.integralfamily()->name()//
				<< equiv.id()//
				<< " [color=green];\n";
	}
	for (s = secs.begin(); s != secs.end(); ++s) {
		set<Sector> sub = get_reduced_subsectors(*s, true);
		allsecs.insert(sub.begin(), sub.end());
	}

	// sectors are vertices, relations are directed edges
	map<Sector, set<Sector> > relations;
	for (s = allsecs.begin(); s != allsecs.end(); ++s)
		relations[*s] = get_reduced_subsectors(*s, false);

	map<Sector, set<Sector> >::const_iterator r;
	for (r = relations.begin(); r != relations.end(); ++r) {
		for (s = r->second.begin(); s != r->second.end(); ++s)
			file//
					<< r->first.integralfamily()->name() << r->first.id()//
					<< " -> "//
					<< s->integralfamily()->name()//
					<< s->id()//
					<< " [color=blue];\n";
	}
	for (s = allsecs.begin(); s != allsecs.end(); ++s)
		file << s->integralfamily()->name()//
				<< s->id()//
				<< " [label=\"" << *s
		/*<< " " << s->t()*/<< "\"];\n";

	// trailer
	file << "}\n";

	file.close();
	rename(filename_tmp, filename);
}

void operator >>(const YAML::Node& node, SectorMappings& m) {
	using namespace YAML;
	using namespace GiNaC;
	string name;
	node["name"] >> name;
	string required = m.integralfamily_->name();
	if (name != required)
		throw runtime_error("Name " + name + " does not match " + required);

	const Node& z_node = node["zero_sectors"];
	for (Iterator n = z_node.begin(); n != z_node.end(); ++n)
		for (Iterator s = n.second().begin(); s != n.second().end(); ++s)
			m.zero_sectors_.insert(*s);

	if (node.FindValue("sectors_without_graph")) {
		const Node& ng_node = node["sectors_without_graph"];
		for (Iterator n = ng_node.begin(); n != ng_node.end(); ++n)
			for (Iterator s = n.second().begin(); s != n.second().end(); ++s)
				m.sectors_without_graph_.insert(*s);
	}

	const Node& rel_node = node["sector_relations"];
	for (Iterator r = rel_node.begin(); r != rel_node.end(); ++r) {
		int from_equiv_id;
		r.first() >> from_equiv_id;
		Sector from_equiv_sec(m.integralfamily(), from_equiv_id);
		Sector to_equiv_sec(r.second()[0]);

		bool new_notation = true;
		if (r.second().size() == 3) {
			new_notation = false;
			int from_id;
			r.second()[1] >> from_id;
			Sector from_sec(m.integralfamily(), from_id);
			VERIFY(from_sec == from_equiv_sec);
		} else {
			VERIFY(r.second().size() == 2);
		}

		const Node& shift_node = new_notation ? r.second()[1] : r.second()[2];
		exmap shift;
		lst from_momenta = m.integralfamily()->get_all_momenta();
		lst to_momenta = to_equiv_sec.integralfamily()->get_all_momenta();
		for (Iterator s = shift_node.begin(); s != shift_node.end(); ++s) {
			string lhs, rhs;
			(*s)[0] >> lhs;
			(*s)[1] >> rhs;
			shift[ex(lhs, from_momenta)] = ex(rhs, to_momenta);
		}
		if (!SectorMappings::test_shift(from_equiv_sec, to_equiv_sec, shift))
			ABORT("Wrong shift from sector " << from_equiv_sec << " to sector "
					<< to_equiv_sec << " detected: " << shift);
		m.sector_relations_.insert(make_pair(from_equiv_id, to_equiv_sec));
		m.sector_relation_shifts_.insert(make_pair(from_equiv_id, shift));
	}

	if (node.FindValue("sector_symmetries")) {
		const Node& sym_node = node["sector_symmetries"];
		const lst& symbols = m.integralfamily_->get_all_momenta();
		for (Iterator r = sym_node.begin(); r != sym_node.end(); ++r) {
			int id;
			r.first() >> id;
			Sector sector(m.integralfamily_, id);
			list<exmap> id_shifts;
			for (Iterator s = r.second().begin(); s != r.second().end(); ++s) {
				exmap shift;
				read(*s, shift, symbols);
				YAML::Emitter ye;
				ye << Flow << shift;
				if (!SectorMappings::test_shift(sector, sector, shift))
					ABORT("Wrong symmetry shift for topology " << id << " detected: " << ye.c_str());
				id_shifts.push_back(shift);
			}
			m.sector_symmetries_.insert(make_pair(id, id_shifts));
		}
	}
}

YAML::Emitter & operator<<(YAML::Emitter& os, const SectorMappings& m) {
	using namespace YAML;
	using namespace GiNaC;
	const IntegralFamily* ic = m.integralfamily();
	os << BeginMap;

	os << Key << "name" << Value << ic->name();

	os << Key << "zero_sectors" << Value << BeginMap;
	map<int, set<int> > z_by_t; // collect by common t
	set<int>::const_iterator z;
	for (z = m.zero_sectors_.begin(); z != m.zero_sectors_.end(); ++z)
		z_by_t[ic->get_t(*z)].insert(*z);
	map<int, set<int> >::iterator i;
	for (i = z_by_t.begin(); i != z_by_t.end(); ++i) {
		stringstream ss;
		ss << "t=" << i->first;
		os << Key << ss.str() << Value << Flow << BeginSeq;
		for (set<int>::iterator s = i->second.begin(); s != i->second.end(); ++s)
			os << *s;
		os << EndSeq;
	}
	os << EndMap;

	os << Key << "sectors_without_graph" << Value << BeginMap;
	map<int, set<int> > ng_by_t; // collect by common t
	set<int>::const_iterator seci;
	for (seci = m.sectors_without_graph_.begin(); seci
			!= m.sectors_without_graph_.end(); ++seci)
		ng_by_t[ic->get_t(*seci)].insert(*seci);
	//map<int, set<int> >::iterator i;
	for (i = ng_by_t.begin(); i != ng_by_t.end(); ++i) {
		stringstream ss;
		ss << "t=" << i->first;
		os << Key << ss.str() << Value << Flow << BeginSeq;
		for (set<int>::iterator s = i->second.begin(); s != i->second.end(); ++s)
			os << *s;
		os << EndSeq;
	}
	os << EndMap;

	os << Key << "sector_relations" << Value << BeginMap;
	const map<int, exmap>& shifts = m.sector_relation_shifts_;
	map<int, GiNaC::exmap>::const_iterator r;
	for (r = shifts.begin(); r != shifts.end(); ++r) {
		int from_equiv = r->first;
		// const Sector& from_equiv = m.get_equivalent(from);
		const exmap& shift = r->second;
		map<int, Sector>::const_iterator equiv;
		equiv = m.sector_relations_.find(from_equiv);
		if (equiv == m.sector_relations_.end())
			ERROR("incomplete sector relations");
		const Sector& to_equiv = equiv->second;
		os << Key << from_equiv << Value << Flow << BeginSeq;
		os << to_equiv;
		//VERIFY(from_equiv.id() == from.id());
		// os << from.id();
		os << BeginSeq;
		for (exmap::const_iterator sr = shift.begin(); sr != shift.end(); ++sr)
			os << BeginSeq << sr->first << sr->second << EndSeq;
		os << EndSeq;
		os << EndSeq;
	}
	os << EndMap;

	os << Key << "sector_symmetries" << Value << BeginMap;
	const map<int, list<GiNaC::exmap> >& syms = m.sector_symmetries_;
	map<int, list<GiNaC::exmap> >::const_iterator it;
	list<GiNaC::exmap>::const_iterator it2;
	for (it = syms.begin(); it != syms.end(); ++it) {
		os << Key << it->first << Value;
		os << BeginSeq;
		for (it2 = it->second.begin(); it2 != it->second.end(); ++it2) {
			os << Flow << *it2;
		}
		os << EndSeq;
	}
	os << EndMap;

	os << EndMap;
	return os;
}

} // namespace Reduze
